#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCoordinate4.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/draggers/SoDragPointDragger.h>

SoSeparator *ParticleRoot;
SoSeparator *ScreenRoot;
SoSeparator *EmitterRoot;
SoSeparator *DraggerRoot;

SbVec3f *VectorFieldCoord;
SoCoordinate3 *StreamlineCoord;
SbVec3f EmitterPoint(0.0, 0.0, 0.0);
SoXtExaminerViewer *myViewer;

float PresentTime = 0.0;
float DTime = 0.01;
int numParticles = 0;

const float screenRadius   = 2.0;
const float sphereDistance = 14.0;
const int   numSteps       = 25;
const int   maxParticles   = 40;
const int   textureXDimI   = 512;
const int   textureYDimI   = textureXDimI;
const double textureXDimF   = (double) textureXDimI;
const double textureYDimF   = textureXDimF;
const int   numComponents  = 3;

const SbVec3f draggerDif = SbVec3f(screenRadius, screenRadius, screenRadius);
SbVec2s imgSize = SbVec2s(textureXDimI, textureYDimI);
unsigned char myImageBuffer[textureXDimI*textureYDimI*numComponents] = {0};

//
// Kevin Beason
// 3/1/00
//

void getPointer(double &a, double &b, double &c, int x, int y){
  double phi, theta, t;

  phi = (double)y/(textureYDimF-1.0)*M_PI;
  theta = (double)x/(textureXDimF-1.0)*2.0*M_PI;

  b=cos(M_PI-phi);
  t=tan(3.0/2.0*M_PI-theta);
  a=sqrt((1.0-b*b)/(1.0+t*t));
  if (theta<M_PI && theta>=0.0) a=-a;
  c=a*t;
  
  return;
}

void getXY (int &x, int &y, double a, double b, double c){
  double phi, theta;
  phi=M_PI-acos(b/sqrt(a*a+b*b+c*c));
  theta=3.0*M_PI/2.0-atan2(c, a);
  while (theta< 0.0     ) theta+=2.0*M_PI;
  while (theta>=2.0*M_PI) theta-=2.0*M_PI;
  
  // step 2
  // determine texture coordinates
  y=(int)((textureYDimF-0.999) * phi/M_PI);
  x=(int)((textureXDimF-0.999) * theta/(2.0*M_PI));

  if (a>0.0 && x==0) x=textureXDimI-1;;
  
  return;
}

void drawDot (SbVec3f pointer, double r, double g, double b, int dx=1, int dy=1){

  // step 1
  // determine phi and theta
  double phi, theta;
  int x, y, index;
  phi=M_PI-acos((double)pointer[1]/(double)pointer.length());
  theta=3.0*M_PI/2.0-atan2((double)pointer[2], (double)pointer[0]);
  while (theta< 0.0     ) theta+=2.0*M_PI;
  while (theta>=2.0*M_PI) theta-=2.0*M_PI;

  // step 2
  // determine texture coordinates
  y=(int)((textureYDimF-0.999) * phi/M_PI);
  x=(int)((textureXDimF-0.999) * theta/(2.0*M_PI));

  // hack
  if (pointer[0]>0.0 && x==0) x=textureXDimI-1;
  
  // step 3
  // vary x and y
  
  // step 4
  // draw dot
  int i, j;
  int l, m;
  for (i=0;i<dx;i++){
    for (j=0;j<dy;j++){
      l=x+i;
      if (l>=textureXDimI) continue;
      m=y+j;
      if (m>=textureYDimI) continue;
      index = textureXDimI*numComponents*m + numComponents*l;
      myImageBuffer[index+0] = (int)(r*255.0);
      myImageBuffer[index+1] = (int)(g*255.0);;
      myImageBuffer[index+2] = (int)(b*255.0);;
    }
  }
      
  return;
}

void updateTexture(){
  SoTexture2 *screenTexture2;
  screenTexture2 = (SoTexture2 *) ScreenRoot->getChild(2);
  screenTexture2->image.setValue(imgSize, numComponents, myImageBuffer);
}

int dx=textureXDimI;
int dy=textureYDimI;
int counter=0;
int row=0;
int col=0;
SbVec3f old_offset;

SoSeparator *
makeItMove (){

  SoSeparator *currParticleSep;
  SoTransform *currParticleTrans;
  double radius, discrim;
  double a, b, c, dist, temp;
  double closest_dist;
  int closest_intersecting_ball;
  SoMaterial *currMaterial;
  SbColor color;
  SbVec3f ray, ball;
  SbVec3f offset;

  SoTransform *screenTrans = (SoTransform *)ScreenRoot->getChild(0);
  offset = ((SoDragPointDragger *)DraggerRoot->getChild(0))->translation.getValue() - draggerDif;

  // move and redraw screen if dragger has moved
  if (offset!=old_offset){
    screenTrans->translation.setValue(offset);
    dx=textureXDimI;
    dy=textureYDimI;
    counter=0;
    row=0;
    col=0;
  }
  old_offset=offset;

  if (dx==0 && dy==0) return ParticleRoot;  
  
  while (1){
    while (1){
      //  for (col=0;col<textureXDimI;col+=dx){
      //    for (row=0;row<textureYDimI;row+=dy){
  
      // pick a random ray
      //ray = SbVec3f(drand48()-0.5, drand48()-0.5, drand48()-0.5);
      //ray.normalize();
      double ma, mb, mc;
      getPointer(ma, mb, mc, row, col);
      ray.setValue(ma, mb, mc);

      // loop through circles
      closest_dist=1000.0;
      closest_intersecting_ball=-1;
  
      int i;
      for (i=0; i<numParticles; i++){
	currParticleSep = (SoSeparator *) ParticleRoot->getChild(i);
	currParticleTrans = (SoTransform *) currParticleSep->getChild(1);
	ball = currParticleTrans->translation.getValue() - offset;

	if (ball.dot(ray)<0.0) continue;
    
	radius = currParticleTrans->scaleFactor.getValue()[0];

	// solve quadratic forumla to determine if
	// there is a ball intersection
	a = 1.0;
	b = -2.0*(ball[0]*ray[0]+ball[1]*ray[1]+ball[2]*ray[2]);
	c = pow(ball.length(), 2.0) - radius*radius;
	discrim = b*b - 4.0*a*c;

	if (discrim>0.0){
	  discrim = sqrt(discrim);
	  temp = (-b+discrim)/(2.0*a);
	  dist = (-b-discrim)/(2.0*a);
	  if (temp<0.0) temp=-temp;
	  if (dist<0.0) dist=-dist;
	  if (temp<dist) dist=temp;
	  if (dist<closest_dist){
	    closest_dist=dist;
	    closest_intersecting_ball=i;
	  }
	}
      
      }

  
      // if intersection draw dot
      // dont forget to implement sorting later
      if (closest_intersecting_ball!=-1){
	currParticleSep =
	  (SoSeparator *) ParticleRoot->getChild(closest_intersecting_ball);
	currMaterial = (SoMaterial *) currParticleSep->getChild(0);
	color=currMaterial->diffuseColor[0];
	currParticleTrans = (SoTransform *) currParticleSep->getChild(1);
	ball=currParticleTrans->translation.getValue() - offset;
	ball=ball-ray*closest_dist;
	ball.normalize();
	double angle = ball.dot(ray);
    
	drawDot(ray, color[0]*angle, color[1]*angle, color[2]*angle, dx, dy);
      }
      else {
	drawDot(ray, 0.3, 0.3, 0.3, dx, dy);
      }
      
//for (col=0;col<textureXDimI;col+=dx){
//  for (row=0;row<textureYDimI;row+=dy){
      counter++;
      row+=dy;
      if (row>=textureYDimI){
        row=0;
	break;
      }
      if (counter>255){
        counter=0;
        updateTexture();
        return ParticleRoot;
      }
    }

    
    col+=dx;
    if (col>=textureXDimI){
      dx=dx/2;
      dy=dy/2;
      col=0;
    }

    if (counter>255){
      counter=0;
      updateTexture();
      return ParticleRoot;
    }
  }

  counter=0;
  updateTexture();
  return ParticleRoot;
}

SoSeparator *
makeScene ()
  {
  SoSeparator *root = new SoSeparator();
  srand48(0);

  

  
  SoShapeHints *twoSided = new SoShapeHints();
  twoSided->shapeType = SoShapeHints::UNKNOWN_SHAPE_TYPE;
  twoSided->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
  root->addChild(twoSided);

  ////////////////////////////////////////////////////////////
  //
  // Create the screen
  //
  //                      ScreenRootSep
  //                          |
  //          ---------------------------------
  //          |      |           |            |
  //          0      1           2            3
  //          |      |           |            |
  //      transform complexity texture       sphere

  // create screen root
  SoSeparator *screenRootSep = new SoSeparator();
  ScreenRoot = screenRootSep;
  root->addChild(screenRootSep);

  // create screen transform
  SoTransform *screenTransform = new SoTransform();
  screenTransform->translation.setValue(0.0, 0.0, 0.0);
  screenTransform->scaleFactor.setValue(1.0, 1.0, 1.0);
  screenRootSep->addChild( screenTransform );

  // create screen complexity
  SoComplexity *screenComplexity = new SoComplexity();
  screenComplexity->type.setValue(SoComplexity::SCREEN_SPACE);
  screenComplexity->value.setValue(0.7);
  screenComplexity->textureQuality.setValue(1.0);
  screenRootSep->addChild( screenComplexity );
  
  // create screen texture
  SoTexture2 *screenTexture2 = new SoTexture2();
  screenRootSep->addChild( screenTexture2 );
  int i;
  for (i=0;i<textureXDimI*textureYDimI*numComponents;i++) myImageBuffer[i]=255;
  screenTexture2->image.setValue(imgSize, numComponents, myImageBuffer);

  // sreate screen (sphere)
  SoSphere *screenSphere = new SoSphere();
  screenSphere->radius.setValue(screenRadius);
  screenRootSep->addChild( screenSphere );

  ////////////////////////////////////////////////////////////
  //
  //                  ParticleRootSep
  //                        |
  //          ------------------------------
  //          |       |      |  |  |  |  ...
  //          0       1      2  3  ...
  //                         |
  //                   ParticleSep[2]
  //                         |
  //                      -------
  //                      |     |
  //                 transform sphere

  SoSeparator *particleRootSep = new SoSeparator();
  ParticleRoot = particleRootSep;
  root->addChild(particleRootSep);

  static int currentParticleNum;
  static SbVec3f direction;
  double distance, size;
  double r, g, b;
  SoSeparator *currentParticleSep;

  for (i=0;numParticles<maxParticles;i++){
  
  ////////////////////////////////////////////////////////////////////////
  //  
  // either create a new sphere if we're not already working with one
  // set pointer to particleSep in either case
  //
  //                   ParticleSep[0]
  //                         |
  //                      ------------------
  //                      |     |          |
  //                 material transform sphere
  //

  currentParticleNum=numParticles;
  numParticles++;
  currentParticleSep = new SoSeparator();
  ParticleRoot->addChild( currentParticleSep );
    
  SoTransform *particleTransform = new SoTransform();

  // set random color
  r = drand48();
  g = drand48();
  b = drand48();
  SoMaterial *particleMaterial = new SoMaterial();
  particleMaterial->diffuseColor.setValue(r, g, b);
  currentParticleSep->addChild( particleMaterial );               

  // set random size and position
  size = drand48()/2.0 + 3.0;
  direction = SbVec3f(drand48()-0.5, drand48()-0.5, drand48()-0.5);
  direction.normalize();
  distance = drand48() * sphereDistance + screenRadius + size + 5.0;
  particleTransform->translation.setValue(direction*distance);
  particleTransform->scaleFactor.setValue(size, size, size);
  currentParticleSep->addChild( particleTransform );

  // add sphere
  SoSphere *particleSphere = new SoSphere();
  particleSphere->radius.setValue(1.0);
  currentParticleSep->addChild( particleSphere );

  // find pointer
  SbVec3f pointer = particleTransform->translation.getValue();
  pointer = pointer - screenTransform->translation.getValue();
  pointer.normalize();

  // draw dot
  // drawDot(pointer, r, g, b );

  }

  // add sphere dragger
  SoDragPointDragger *sphereDrag = new SoDragPointDragger();
  sphereDrag->translation.setValue(3.0, 3.0, 3.0);
  DraggerRoot = new SoSeparator();
  root->addChild(DraggerRoot);
  DraggerRoot->addChild(sphereDrag);

  return root;
  }


void
animate(void *, SoSensor *)
  {
  makeItMove ();
  PresentTime += DTime;
  }

void
main(int, char **argv)
  {
  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) exit(1);

  SoSeparator *root = makeScene();

  SoTimerSensor *sensor = new SoTimerSensor(animate, root);
  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->schedule();

  myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setTransparencyType(SoGLRenderAction::BLEND);
  myViewer->setBackgroundColor(SbColor(0.3, 0.3, 0.3));
  myViewer->setSceneGraph(root);
  myViewer->setTitle("Emitter");
  myViewer->show();

  SoXt::show(myWindow);
  SoXt::mainLoop();
  }
